from pathlib import Path
from typing import Tuple, Union
from beastiary import crud, schemas
from beastiary.models.trace import Trace
from sqlalchemy.orm.session import Session
import errno
import os


def get_headers(path: Path) -> Tuple[int, str]:
    with open(path, "r") as f:
        headers_set = False
        while True:
            line = f.readline()
            if line.startswith("#"):
                continue
            headers_list = line.split()
            try:
                headers_list[0] = "state"
            except IndexError:
                raise ValueError(f"Could not find headers in {trace_in.path}")
            headers_line = " ".join(headers_list)
            last_byte = f.tell()
            return last_byte, headers_line


def add_trace(db: Session, trace_in: schemas.TraceCreate) -> Trace:
    if not trace_in.path.is_file():
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), trace_in.path)
    last_byte, headers_line = get_headers(trace_in.path)
    trace = crud.trace.create(
        db=db, obj_in=trace_in, headers_line=headers_line, last_byte=last_byte
    )
    return trace



from os import error, path
from typing import Any, Optional, Callable
from fastapi import FastAPI
from fastapi import APIRouter, Request
from fastapi.logger import logger
from fastapi.staticfiles import StaticFiles
from starlette.responses import JSONResponse, FileResponse
from starlette.middleware.cors import CORSMiddleware
from starlette.datastructures import URL

import logging

from beastiary.api.endpoints import traces, samples


class BeastiaryAPI(FastAPI):
    def __init__(
        self,
        title: str = "Beastiary",
        description: str = "Realtime and remote trace inspection",
        version: str = "0.8.2",
    ) -> None:
        super().__init__(
            title=title, description=description, version=version
        )  # Initialise FastAPI super class
        self.token: Optional[str] = None
        self.security: bool = True


api = BeastiaryAPI()

gunicorn_logger = logging.getLogger("gunicorn.error")
logger.handlers = gunicorn_logger.handlers
logger.setLevel(gunicorn_logger.level)


file_path = path.abspath(path.dirname(__file__))
webapp_path = path.join(file_path, "../webapp-dist/")
index_path = path.join(file_path, "../webapp-dist/index.html")


@api.get("/")
async def index() -> FileResponse:
    return FileResponse(index_path)


@api.get("/api/security/token", tags=["security"])
async def test_token() -> Any:
    return {"token": api.token}


api_router = APIRouter(prefix="/api")
api_router.include_router(samples.router, prefix="/samples", tags=["samples"])
api_router.include_router(traces.router, prefix="/traces", tags=["traces"])
api.include_router(api_router)

api.mount("/", StaticFiles(directory=webapp_path))

# replace with https://tiangolo.medium.com/nice-a6eafd9a7bca
# probably need some switch to run it off
# auto_error: bool = False
@api.middleware("http")
async def auth_check(request: Request, call_next: Callable) -> Any:
    if "/api/" in request.url.path and request.app.security == True:
        token = request.headers.get("Authorization")
        if not token or len(token.split()) < 2:
            return JSONResponse(
                content={"detail": "Authorization header not provided!"},
                status_code=401,
            )
        else:
            token = token.split()[1]
        if token != request.app.token:
            return JSONResponse(content={"detail": "Invalid token!"}, status_code=401)
    response = await call_next(request)
    return response


@api.middleware("http")
async def add_custom_header(request: Request, call_next: Callable) -> Any:
    response = await call_next(request)
    if response.status_code == 404 and "/api/" not in request.url.path:
        return FileResponse(index_path)
    return response


api.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)



from beastiary.api.core import add_trace
from beastiary.models.trace import Trace
from beastiary.schemas import sample
from typing import Any, List
from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from beastiary import crud, schemas
from beastiary.api import deps

router = APIRouter()


@router.get("/", response_model=List[schemas.Trace])
def get_traces(
    db: Session = Depends(deps.get_db),
    skip: int = 0,
    limit: int = 100,
) -> List[Trace]:
    """
    Retrieve traces.
    """
    traces = crud.trace.get_multi(db=db, skip=skip, limit=limit)
    return traces


@router.get("/{trace_id}", response_model=schemas.Trace)
def get_trace(trace_id: int, db: Session = Depends(deps.get_db)) -> Trace:
    """
    Retrieve traces.
    """
    trace = crud.trace.get(db=db, id=trace_id)
    if not trace:
        raise HTTPException(404, "Trace not found!")
    return trace


@router.post("/", response_model=schemas.Trace)
def create_trace(
    *,
    db: Session = Depends(deps.get_db),
    trace_in: schemas.TraceCreate,
) -> Trace:
    """
    Create new trace.
    """
    try:
        trace = add_trace(db, trace_in)
    except FileNotFoundError as e:
        raise HTTPException(404, detail="Could not find log file!")
    except Exception as e:
        raise HTTPException(500, detail=f"Could not add {trace_in.path}")
    # get samples
    return trace



from typing import Any, List, Tuple
import math
from beastiary.models.trace import Trace
from beastiary.schemas.sample import SampleCreate
from fastapi import APIRouter, Depends, HTTPException, Path
from sqlalchemy.orm import Session

from beastiary import crud, schemas
from beastiary.api import deps
from beastiary.log import logger

router = APIRouter()


def read_lines(trace: Trace) -> Tuple[int, list]:
    if not trace.path:
        raise ValueError("Path must be set.")
    with open(trace.path, "r") as f:
        f.seek(trace.last_byte, 0)
        lines = f.readlines()
        if lines:
            last_byte = f.tell()
        else:
            last_byte = trace.last_byte
        return last_byte, lines


def lines_to_SampleCreate(headers: list, lines: list) -> List[SampleCreate]:
    samples = []
    for line in lines:
        if not line:
            # blank lines are bad...
            # you'll have to have a smart way to handel this with the byte offset
            raise ValueError("Poorly formated line.")
        data = {}
        for header, value in zip(headers, line.split()):
            value = float(value)
            if value.is_integer():
                value = int(value)
            elif math.isnan(value) or math.isinf(value):
                value = None
            data[header] = value
        sample_in = {}
        sample_in["data"] = data
        sample_in["state"] = data["state"]
        samples.append(schemas.sample.SampleCreate(**sample_in))
    return samples


def check_for_new_samples(db: Session, trace: Trace) -> None:
    last_byte, lines = read_lines(trace)
    in_samples = lines_to_SampleCreate(trace.headers_line.split(), lines)
    if in_samples:
        crud.sample.create_multi_with_trace(db, objs_in=in_samples, trace_id=trace.id)
    # update the trace byte
    crud.trace.update(db, db_obj=trace, obj_in={"last_byte": last_byte})


@router.get("/", response_model=List[schemas.Sample])
def get_samples(
    trace_id: int,
    skip: int = 0,
    limit: int = 100,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Retrieve samples.
    """
    trace = crud.trace.get(db, trace_id)
    if not trace:
        raise HTTPException(404, detail="Trace not found!")

    samples = crud.sample.get_multi_by_trace(
        db, trace_id=trace.id, skip=skip, limit=limit
    )

    logger.debug(f"{len(samples)} found, {limit} samples requested")
    if len(samples) < limit:

        logger.debug(f"Checking for new samples in {trace.path}")
        try:
            check_for_new_samples(db, trace=trace)
        except Exception as e:
            raise HTTPException(500, detail=f"Could read samples in {trace.path}")
        # get samples
        samples = crud.sample.get_multi_by_trace(
            db, trace_id=trace.id, skip=skip, limit=limit
        )
    logger.debug(f"Returning {len(samples)} samples")
    return samples



import errno
import os
import pkg_resources
import uuid
import typer
import uvicorn

from pathlib import Path
from typing import List, Optional
from fastapi.params import Security

from beastiary.api.core import add_trace
from beastiary.api import api
from beastiary.db.init_db import init_db
from beastiary.db.session import SessionLocal
from beastiary import schemas


app = typer.Typer()

db = SessionLocal()
init_db(db)


@app.command()
def main(
    log_files: Optional[List[Path]] = typer.Argument(
        None,
        metavar="[LOG_FILE]...",
        help="Optional path to log file(s) to add at start up.",
    ),
    version: bool = typer.Option(
        False, "--version", "-v", help="Display version number."
    ),
    token: str = typer.Option(str(uuid.uuid4()), "--token", "-t"),
    host: str = typer.Option("127.0.0.1", "--host"),
    port: str = typer.Option(5000, "--port"),
    security: bool = typer.Option(True, help="Turn off token requirement."),
    debug: bool = typer.Option(False, help="Set debug mode."),
    testing: bool = typer.Option(False, help="Only for testing.", hidden=True),
) -> None:
    """
    Realtime and remote trace inspection with BEASTIARY.
    """
    if version:
        typer.echo(f"Beastiary {pkg_resources.get_distribution('beastiary').version}")
        return typer.Exit()
    msg = typer.style("STARTING BEASTIARY", fg=typer.colors.BLUE, bold=True)
    typer.echo(f"\n🐙🐁 {msg} 🐁🐙\n")
    if log_files:
        typer.echo(f"Adding log files:")
        for path in log_files:
            try:
                trace = add_trace(db, schemas.TraceCreate(path=str(path)))
                typer.echo(f"✅ - {trace.path}")
            except ValueError:
                typer.echo(f"❌ - {trace.path}")
        typer.echo("")
    url = typer.style(
        f"http://{host}:{port}/login?token={token}", fg=typer.colors.GREEN, bold=False
    )
    typer.echo(f"Go to: {url}\n")
    typer.echo(f"If prompted enter token: {token}")
    setattr(api, "token", token)
    if not security:
        warning = typer.style("WARNING", fg=typer.colors.YELLOW, bold=True)
        typer.echo(f"{warning}: Security disabled!")
        setattr(api, "security", False)
    log_level = "warning"
    if debug:
        log_level = "debug"
    if not testing:
        uvicorn.run(api, host=host, port=port, log_level=log_level)



from typing import List

from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from beastiary.crud.base import CRUDBase
from beastiary.models.sample import Sample
from beastiary.schemas.sample import SampleCreate, SampleUpdate


class CRUDSample(CRUDBase[Sample, SampleCreate, SampleUpdate]):
    def create_with_trace(
        self, db: Session, *, obj_in: SampleCreate, trace_id: int
    ) -> Sample:
        obj_in_data = jsonable_encoder(obj_in)
        db_obj = self.model(**obj_in_data, trace_id=trace_id)
        db.add(db_obj)
        db.commit()
        db.refresh(db_obj)
        return db_obj

    def create_multi_with_trace(
        self, db: Session, *, objs_in: List[SampleCreate], trace_id: int
    ) -> List[Sample]:
        objs_in_data = [jsonable_encoder(obj_in) for obj_in in objs_in]
        for obj_in in objs_in_data:
            obj_in.update({"trace_id": trace_id})
        db_objs = [self.model(**obj_in_data) for obj_in_data in objs_in_data]
        db.add_all(db_objs)
        db.commit()
        return db_objs

    def get_multi_by_trace(
        self, db: Session, *, trace_id: int, skip: int = 0, limit: int = 100
    ) -> List[Sample]:
        return (
            db.query(self.model)
            .filter(Sample.trace_id == trace_id)
            .offset(skip)
            .limit(limit)
            .all()
        )


sample = CRUDSample(Sample)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['beastiary',
 'beastiary.api',
 'beastiary.api.endpoints',
 'beastiary.crud',
 'beastiary.db',
 'beastiary.models',
 'beastiary.schemas']

package_data = \
{'': ['*'],
 'beastiary': ['webapp-dist/*',
               'webapp-dist/css/*',
               'webapp-dist/img/icons/*',
               'webapp-dist/js/*']}

install_requires = \
['SQLAlchemy>=1.4.22,<2.0.0',
 'aiofiles<0.6.0',
 'fastapi[all]>=0.67.0,<0.68.0',
 'typer[all]>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['beastiary = beastiary.cli:app']}

setup_kwargs = {
    'name': 'beastiary',
    'version': '0.9.6',
    'description': '',
    'long_description': '![beastiary logo](https://beastiary.wytamma.com/images/logo.png)\n\n\n[![PyPi](https://img.shields.io/pypi/v/beastiary.svg)](https://pypi.org/project/beastiary/)\n[![tests](https://github.com/Wytamma/beastiary/actions/workflows/test.yml/badge.svg)](https://github.com/Wytamma/beastiary/actions/workflows/test.yml)\n[![cov](https://codecov.io/gh/Wytamma/beastiary/branch/master/graph/badge.svg)](https://codecov.io/gh/Wytamma/beastiary)\n[![docs](https://github.com/Wytamma/beastiary/actions/workflows/docs.yml/badge.svg)](https://beastiary.wytamma.com/)\n\nThis is a real time version of tracer that can be run on remote servers (e.g. a HPC). Its goal is to be a beautiful and simple, yet power tool for Bayesian phylogenetic inference.\n\n## Installation\n```bash\npip install beastiary\n```\n\n## Use\nTo start beastiary use the `beastiary` command. This will start the beastiary server. \n\n```bash\nbeastiary\n```\n\nThe server will start on [http://127.0.0.1:5000/](http://127.0.0.1:5000/) by default. Every time the server starts it will generate a unique id that can be used to log into the webapp. \n\nGo to [http://127.0.0.1:5000/](http://127.0.0.1:5000/) and log in with the ID. \n\n![](docs/images/screen_shot.png)',
    'author': 'Wytamma Wirth',
    'author_email': 'wytamma.wirth@me.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)



from typing import TYPE_CHECKING

from sqlalchemy import Column, Integer, JSON, ForeignKey
from sqlalchemy.orm import relationship

from beastiary.db.base_class import Base

if TYPE_CHECKING:
    from beastiary.models.trace import Trace


class Sample(Base):
    id = Column(Integer, primary_key=True)
    state = Column(Integer)
    data = Column(JSON)
    trace_id = Column(Integer, ForeignKey("trace.id"))
    trace = relationship("Trace", back_populates="samples")



from typing import List

from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from beastiary.crud.base import CRUDBase
from beastiary.models.trace import Trace
from beastiary.schemas.trace import TraceCreate, TraceUpdate


class CRUDTrace(CRUDBase[Trace, TraceCreate, TraceUpdate]):
    def create(  # type: ignore
        self, db: Session, *, obj_in: TraceCreate, headers_line: str, last_byte: int
    ) -> Trace:
        obj_in_data = jsonable_encoder(obj_in)
        db_obj = self.model(
            **obj_in_data, headers_line=headers_line, last_byte=last_byte
        )
        db.add(db_obj)
        db.commit()
        db.refresh(db_obj)
        return db_obj


trace = CRUDTrace(Trace)
